@@ -10,11 +10,13 @@ from __future__ import annotations
 import asyncio
 import functools
 import inspect
+from dataclasses import dataclass
 import logging
 import os
+import threading
 import time
 import uuid
-from typing import Any, Awaitable, Callable, Dict, List, Optional, Sequence, Tuple
+from typing import Any, Awaitable, Callable, Dict, List, Optional, Sequence, Tuple, Union
 from datetime import datetime
 import tiktoken
 
@@ -44,7 +46,24 @@ except ImportError:
     _HAS_RICH = False
     _console = None
 
-__all__ = ["run_simulations", "track"]
+from .webhook_trigger import WebhookTrigger, create_webhook_target_agent
+from .multimodal import MultimodalOutput, FileAttachment
+
+# Type definitions for multimodal support
+MultimodalAgent = Callable[[str], Awaitable[MultimodalOutput] | MultimodalOutput]
+MultimodalAgentFactory = Callable[[], MultimodalAgent]
+
+__all__ = ["run_simulations", "track", "record_tool_event", "start_tool_event", "finish_tool_event", "ToolEventSpanHandle", "WebhookTrigger", "create_webhook_target_agent", "MultimodalOutput", "FileAttachment"]
+
+@dataclass
+class ToolEventSpanHandle:
+    """Handle for managing manual tool event spans.
+    
+    This dataclass stores the span reference and start time for manual tracing
+    operations that span multiple function calls.
+    """
+    span: Any  # OpenTelemetry Span object
+    start_time: float  # Start time from time.perf_counter()
 
 # Configuration
 _log = logging.getLogger(__name__)
@@ -282,6 +301,301 @@ def _capture_function_output(span, result: Any) -> None:
         span.set_attribute("function.output", truncated)
 
 
+def record_tool_event(
+    tool_name: str,
+    tool_input: Any = None,
+    tool_output: Any = None,
+    success: Optional[bool] = None,
+    duration_ms: Optional[float] = None,
+    error: Optional[Union[str, Exception]] = None
+) -> None:
+    """Record a tool event as a single span.
+    
+    This function creates and immediately finishes a span for a tool event,
+    capturing all relevant information in one call.
+    
+    Args:
+        tool_name: Name of the tool being traced
+        tool_input: Input data passed to the tool (will be truncated if too long)
+        tool_output: Output data returned by the tool (will be truncated if too long)
+        success: Whether the tool execution was successful (auto-computed from error if not provided)
+        duration_ms: Duration of the tool execution in milliseconds (auto-computed if not provided)
+        error: Error that occurred during tool execution (string or Exception)
+    
+    Examples:
+        # Basic usage
+        record_tool_event("search_api", "query: weather in NYC", "temperature: 72°F")
+        
+        # With error handling
+        try:
+            result = call_external_api()
+            record_tool_event("api_call", "request", result)
+        except Exception as e:
+            record_tool_event("api_call", "request", error=e)
+        
+        # With custom duration
+        start_time = time.perf_counter()
+        result = process_data()
+        duration = (time.perf_counter() - start_time) * 1000
+        record_tool_event("data_processing", "input_data", result, duration_ms=duration)
+    """
+    if not _HAS_OTEL or not _TRACING_INITIALIZED:
+        return
+    
+    try:
+        tracer = trace.get_tracer(__name__)
+        span_name = f"{tool_name}_operation"
+        
+        # Get parent span BEFORE creating new span
+        parent_span = trace.get_current_span()
+        parent_trace_id = None
+        if parent_span and parent_span.is_recording():
+            parent_trace_id = parent_span.get_span_context().trace_id
+        
+        with tracer.start_as_current_span(span_name) as span:
+            start_time = time.perf_counter()
+            
+            # Set basic function attributes
+            span.set_attribute("function.name", tool_name)
+            
+            # Get conversation context from baggage
+            conv_id = baggage.get_baggage("conv_id")
+            simulation_id = baggage.get_baggage("simulation_id")
+            is_simulation = baggage.get_baggage("janus_simulation") == "true"
+            
+            if conv_id:
+                span.set_attribute("conversation.id", conv_id)
+                span.set_attribute("janus.conversation_id", conv_id)
+            
+            if is_simulation:
+                span.set_attribute("janus.simulation", True)
+            
+            if simulation_id:
+                span.set_attribute("janus.simulation_id", simulation_id)
+            
+            # Link to parent trace using the saved parent_trace_id
+            if parent_trace_id:
+                span.set_attribute("trace.id", f"{parent_trace_id:032x}")
+                span.set_attribute("trace.correlation", True)
+            
+            # Set tool input
+            if tool_input is not None:
+                if isinstance(tool_input, (int, float, bool)):
+                    span.set_attribute("function.input", tool_input)
+                else:
+                    str_input = str(tool_input)
+                    truncated = str_input[:100] + "..." if len(str_input) > 100 else str_input
+                    span.set_attribute("function.input", truncated)
+            
+            # Set tool output
+            if tool_output is not None:
+                if isinstance(tool_output, (int, float, bool)):
+                    span.set_attribute("function.output", tool_output)
+                else:
+                    str_output = str(tool_output)
+                    truncated = str_output[:100] + "..." if len(str_output) > 100 else str_output
+                    span.set_attribute("function.output", truncated)
+            
+            # Compute or use provided duration
+            if duration_ms is None:
+                duration_ms = (time.perf_counter() - start_time) * 1000
+            
+            # Compute success if not provided
+            if success is None:
+                success = error is None
+            
+            # Set outcome attributes
+            span.set_attribute("function.duration_ms", round(duration_ms, 3))
+            span.set_attribute("function.success", success)
+            
+            if error is not None:
+                span.set_attribute("function.error", str(error))
+                span.set_attribute("function.error_type", type(error).__name__ if isinstance(error, Exception) else "str")
+    
+    except Exception as e:
+        # Don't let tracing errors break the application
+        _log.warning(f"Failed to record tool event for {tool_name}: {e}")
+
+
+def start_tool_event(
+    tool_name: str,
+    tool_input: Any = None
+) -> Optional[ToolEventSpanHandle]:
+    """Start a tool event span and return a handle for later completion.
+    
+    This function creates a span for a tool event that will be completed later
+    with finish_tool_event(). This is useful for tracing operations that span
+    multiple function calls or async operations.
+    
+    Args:
+        tool_name: Name of the tool being traced
+        tool_input: Input data passed to the tool (will be truncated if too long)
+    
+    Returns:
+        ToolEventSpanHandle containing the span and start time, or None if tracing is disabled
+    
+    Examples:
+        # Basic usage
+        handle = start_tool_event("database_query", "SELECT * FROM users")
+        # ... do database work ...
+        finish_tool_event(handle, "1000 rows returned")
+        
+        # With error handling
+        handle = start_tool_event("external_api", "api_request")
+        try:
+            result = call_api()
+            finish_tool_event(handle, result)
+        except Exception as e:
+            finish_tool_event(handle, error=e)
+        
+        # Check if tracing is enabled
+        handle = start_tool_event("tool_name", "input")
+        if handle is not None:
+            # Tracing is enabled, proceed with work
+            result = do_work()
+            finish_tool_event(handle, result)
+        else:
+            # Tracing is disabled, just do work without tracing
+            result = do_work()
+    """
+    if not _HAS_OTEL or not _TRACING_INITIALIZED:
+        return None
+    
+    try:
+        tracer = trace.get_tracer(__name__)
+        span_name = f"{tool_name}_operation"
+        
+        # Get parent span BEFORE creating new span
+        parent_span = trace.get_current_span()
+        parent_trace_id = None
+        if parent_span and parent_span.is_recording():
+            parent_trace_id = parent_span.get_span_context().trace_id
+        
+        # Create span (not as current span to avoid interfering with existing context)
+        span = tracer.start_span(span_name)
+        start_time = time.perf_counter()
+        
+        # Set basic function attributes
+        span.set_attribute("function.name", tool_name)
+        
+        # Get conversation context from baggage
+        conv_id = baggage.get_baggage("conv_id")
+        simulation_id = baggage.get_baggage("simulation_id")
+        is_simulation = baggage.get_baggage("janus_simulation") == "true"
+        turn_idx = baggage.get_baggage("turn_idx")
+        
+        if conv_id:
+            span.set_attribute("conversation.id", conv_id)
+            span.set_attribute("janus.conversation_id", conv_id)
+        
+        if is_simulation:
+            span.set_attribute("janus.simulation", True)
+        
+        if simulation_id:
+            span.set_attribute("janus.simulation_id", simulation_id)
+        
+        if turn_idx is not None:
+            span.set_attribute("conversation.turn_idx", int(turn_idx))
+            span.set_attribute("janus.turn_idx", int(turn_idx))
+        
+        # Link to parent trace using the saved parent_trace_id
+        if parent_trace_id:
+            span.set_attribute("trace.id", f"{parent_trace_id:032x}")
+            span.set_attribute("trace.correlation", True)
+        
+        # Set tool input
+        if tool_input is not None:
+            if isinstance(tool_input, (int, float, bool)):
+                span.set_attribute("function.input", tool_input)
+            else:
+                str_input = str(tool_input)
+                truncated = str_input[:100] + "..." if len(str_input) > 100 else str_input
+                span.set_attribute("function.input", truncated)
+        
+        return ToolEventSpanHandle(span=span, start_time=start_time)
+    
+    except Exception as e:
+        # Don't let tracing errors break the application
+        _log.warning(f"Failed to start tool event for {tool_name}: {e}")
+        return None
+
+
+def finish_tool_event(
+    handle: ToolEventSpanHandle,
+    tool_output: Any = None,
+    error: Optional[Union[str, Exception]] = None
+) -> None:
+    """Finish a tool event span started with start_tool_event().
+    
+    This function completes a span for a tool event, setting the final attributes
+    and ending the span. It should be called after the tool execution completes.
+    
+    Args:
+        handle: ToolEventSpanHandle returned from start_tool_event()
+        tool_output: Output data returned by the tool (will be truncated if too long)
+        error: Error that occurred during tool execution (string or Exception)
+    
+    Examples:
+        # Successful completion
+        handle = start_tool_event("data_processing", "input_data")
+        result = process_data()
+        finish_tool_event(handle, result)
+        
+        # Error completion
+        handle = start_tool_event("api_call", "request")
+        try:
+            result = call_api()
+            finish_tool_event(handle, result)
+        except Exception as e:
+            finish_tool_event(handle, error=e)
+        
+        # Safe handling of None handle (when tracing is disabled)
+        handle = start_tool_event("tool_name", "input")
+        finish_tool_event(handle, "output")  # Safe even if handle is None
+    """
+    if handle is None:
+        return
+    
+    try:
+        # Validate handle has required attributes
+        if not hasattr(handle, 'span') or not hasattr(handle, 'start_time'):
+            _log.warning("Invalid ToolEventSpanHandle provided to finish_tool_event")
+            return
+        
+        span = handle.span
+        start_time = handle.start_time
+        
+        # Calculate duration
+        duration_ms = (time.perf_counter() - start_time) * 1000
+        
+        # Set tool output
+        if tool_output is not None:
+            if isinstance(tool_output, (int, float, bool)):
+                span.set_attribute("function.output", tool_output)
+            else:
+                str_output = str(tool_output)
+                truncated = str_output[:100] + "..." if len(str_output) > 100 else str_output
+                span.set_attribute("function.output", truncated)
+        
+        # Compute success
+        success = error is None
+        
+        # Set outcome attributes
+        span.set_attribute("function.duration_ms", round(duration_ms, 3))
+        span.set_attribute("function.success", success)
+        
+        if error is not None:
+            span.set_attribute("function.error", str(error))
+            span.set_attribute("function.error_type", type(error).__name__ if isinstance(error, Exception) else "str")
+        
+        # End the span
+        span.end()
+    
+    except Exception as e:
+        # Don't let tracing errors break the application
+        _log.warning(f"Failed to finish tool event: {e}")
+
+
 def _create_context_aware_agent(
     original_agent_factory: Callable, 
     conv_id: str, 
@@ -294,9 +608,10 @@ def _create_context_aware_agent(
     def enhanced_agent_factory():
         original_agent = original_agent_factory()
         
-        async def context_injected_agent(prompt: str) -> str:
+        async def context_injected_agent(prompt: str) -> MultimodalOutput:
             if not _HAS_OTEL:
-                return await _maybe_await(original_agent(prompt, **(persona_kwargs or {})))
+                raw_output = await _maybe_await(original_agent(prompt, **(persona_kwargs or {})))
+                return _normalize_agent_output(raw_output)
             
             # Save current context
             previous_ctx = otel_context.get_current()
@@ -330,15 +645,30 @@ def _create_context_aware_agent(
                         span.set_attribute("janus.turn_idx", int(turn_idx))
                     
                     # Execute agent
-                    result = await _maybe_await(
+                    raw_output = await _maybe_await(
         original_agent(prompt, **(persona_kwargs or {}))
     )
-                    # Record response
-                    truncated_response = result[:200] + "..." if len(result) > 200 else result
+                    
+                    # Normalize output to MultimodalOutput
+                    multimodal_output = _normalize_agent_output(raw_output)
+                    
+                    # Record response (use text for span attributes)
+                    response_text = multimodal_output.to_string()
+                    truncated_response = response_text[:200] + "..." if len(response_text) > 200 else response_text
                     span.set_attribute("agent.response", truncated_response)
                     span.set_attribute("agent.success", True)
                     
-                    return result
+                    # Add multimodal metadata to span
+                    if multimodal_output.text:
+                        span.set_attribute("agent.output.text_length", len(multimodal_output.text))
+                    if multimodal_output.json_data:
+                        span.set_attribute("agent.output.has_json", True)
+                    if multimodal_output.files:
+                        span.set_attribute("agent.output.file_count", len(multimodal_output.files))
+                    if multimodal_output.metadata:
+                        span.set_attribute("agent.output.has_metadata", True)
+                    
+                    return multimodal_output
             finally:
                 # Always detach context to prevent bleed
                 otel_context.detach(token)
@@ -456,9 +786,19 @@ async def _maybe_await(value):
     return value
 
 
+def _normalize_agent_output(output: Union[str, MultimodalOutput]) -> MultimodalOutput:
+    """Convert string output to MultimodalOutput for backward compatibility"""
+    if isinstance(output, str):
+        return MultimodalOutput.from_string(output)
+    elif isinstance(output, MultimodalOutput):
+        return output
+    else:
+        return MultimodalOutput.from_string(str(output))
+
+
 async def arun_simulations(
     *,
-    target_agent: Callable[[], Callable[[str], Awaitable[str] | str]],
+    target_agent: Union[MultimodalAgentFactory, Callable[[], Callable[[str], Awaitable[str] | str]]],
     api_key: str,
     num_simulations: int = 10,
     max_turns: int = 10,
@@ -523,8 +863,8 @@ async def arun_simulations(
         timeout=Timeout(3600),
     )
 
+    progress_ctx: Optional[dict] = None      # ← initialise early
     try:
-        # Set up progress tracking
         progress_ctx = _setup_progress_tracking(num_simulations)
         
         async def _run_single_simulation(sim_idx: int) -> dict:
@@ -564,15 +904,18 @@ async def arun_simulations(
                     for turn_idx in range(max_turns):
                         answer_start_timestamp = datetime.utcnow()
                         answer_start = time.perf_counter()
-                        answer = await _maybe_await(agent(question))
+                        multimodal_answer = await _maybe_await(agent(question))
                         answer_end = time.perf_counter()
                         answer_end_timestamp = datetime.utcnow()
                         answer_generation_time = answer_end - answer_start
                         
+                        # Convert to string for question generation (backward compatibility)
+                        answer_text = multimodal_answer.to_string()
+                        
                         # Track next question timing
                         next_question_start_timestamp = datetime.utcnow()
                         next_question_start = time.perf_counter()
-                        next_question = await client.turn(conv_id, answer)
+                        next_question = await client.turn(conv_id, answer_text)
                         next_question_end = time.perf_counter()
                         next_question_end_timestamp = datetime.utcnow()
                         question_generation_time = next_question_end - next_question_start
@@ -580,7 +923,7 @@ async def arun_simulations(
                         qa_pairs.append({
                             "idx": turn_idx,
                             "q": question,
-                            "a": answer,
+                            "a": multimodal_answer.to_dict(),  # Store full multimodal data
                             "agent_metrics": {
                                 "question_start_timestamp": next_question_start_timestamp.isoformat(),
                                 "question_end_timestamp": next_question_end_timestamp.isoformat(),
@@ -589,7 +932,14 @@ async def arun_simulations(
                                 "question_generation_time": question_generation_time,
                                 "answer_generation_time": answer_generation_time,
                                 "question_tokens": _count_tokens(question),
-                                "answer_tokens": _count_tokens(answer),
+                                "answer_tokens": _count_tokens(answer_text),
+                                "multimodal_metadata": {
+                                    "has_text": multimodal_answer.text is not None,
+                                    "has_json": multimodal_answer.json_data is not None,
+                                    "has_files": multimodal_answer.files is not None,
+                                    "file_count": len(multimodal_answer.files) if multimodal_answer.files else 0,
+                                    "has_metadata": multimodal_answer.metadata is not None
+                                }
                             }
                         })
                         
@@ -639,7 +989,9 @@ async def arun_simulations(
 
 def _setup_progress_tracking(num_simulations: int):
     """Set up Rich progress tracking if available."""
-    if not _HAS_RICH:
+    # Rich's Live() may only run in the main thread / main TTY.
+    if (not _HAS_RICH
+            or threading.current_thread() is not threading.main_thread()):
         return None
         
     progress = Progress(
@@ -785,7 +1137,7 @@ async def _get_hallucination_metrics(
 
 def run_simulations(
     *,
-    target_agent: Callable[[], Callable[[str], Awaitable[str] | str]],
+    target_agent: Union[MultimodalAgentFactory, Callable[[], Callable[[str], Awaitable[str] | str]]],
     api_key: str,
     num_simulations: int,
     max_turns: int,
